/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "FigureFrame.h"

#include "Command.h"
#include "ContextMenu.h"
#include "Debug.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "InformationDialog.h"
#include "TreeView.h"
#include "XmlOptions.h"

#include <QHeaderView>

//_____________________________________________________________
FigureFrame::FigureFrame( QTabWidget* parent ):
    TabWidget( parent )
{

    Debug::Throw("FigureFrame::FigureFrame.\n" );
    setTitle( tr( "Figures" ) );

    setLayout( new QVBoxLayout() );
    layout()->setSpacing(0);
    layout()->setMargin(5);

    layout()->addWidget( list_ = new TreeView( this ) );
    list_->setModel( &model_ );
    list_->setDragEnabled( true );
    list_->setSortingEnabled( false );
    list_->setSelectionMode( QAbstractItemView::ContiguousSelection );
    list_->header()->hide();
    list_->setMask( 1 << FileRecordModel::Filename );

    _installActions();
    QMenu* menu( new ContextMenu( list_ ) );
    menu->addAction( viewFigureAction_ );
    menu->addAction( reloadAction_ );

    // connections
    connect( list_->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), SLOT(_updateActions()) );
    connect( list_, SIGNAL(activated(QModelIndex)), viewFigureAction_, SIGNAL(triggered()) );

}

//_____________________________________________________________
void FigureFrame::setFile( const File& value )
{
    Debug::Throw() <<  "FigureFrame::setFile - " << value << endl;
    if( file_ != value )
    {
        file_ = value;
        requestUpdate();
    }
}

//_____________________________________________________________
void FigureFrame::requestUpdate( void )
{ emit commandRequested( LatexCommand( LatexCommand::FindFigures ) << file_ ); }

//_____________________________________________________________
void FigureFrame::_updateActions( void )
{
    Debug::Throw( "FigureFrame::_updateActions.\n" );
    viewFigureAction_->setEnabled(  !list_->selectionModel()->selectedRows().empty() );
    return;
}

//_____________________________________________________________
void FigureFrame::_viewFigure( void )
{
    Debug::Throw( "FigureFrame::_viewFigure.\n" );

    // retrieve selection
    const File::List selection( model_.filenames( list_->selectionModel()->selectedRows() ) );
    if( selection.empty() )
    {
        InformationDialog( this, tr( "No file selected. <View Figure> canceled." ) ).exec();
        return;
    }

    // check image viewer command
    if( !XmlOptions::get().contains( "IMAGE_VIEWER" ) )
    {
        InformationDialog( this, tr( "Image viewer not set." ) ).exec();
        return;
    }

    for( const auto& file:selection )
    {

        // if file does not exists try append ".eps" extension
        if( !file.exists() )
        {
            const QString buffer = QString( tr( "File '%1' does not exist. <View Figure> canceled." ) ).arg( file );
            InformationDialog( this, buffer ).exec();
            continue;
        }

        // check if file is a directory
        if( file.isDirectory() )
        {
            const QString buffer = QString( tr( "File '%1' is a directory. <View Figure> canceled." ) ).arg( file );
            InformationDialog( this, buffer ).exec();
            continue;
        }

        // update display and run command
        emit messageAvailable( QString( tr( "Running %1..." ) ).arg( QString( XmlOptions::get().raw( "IMAGE_VIEWER" ) ) ) );
        ( Command( XmlOptions::get().raw( "IMAGE_VIEWER" ) ) << file.localName() ).run(file.path());

    }

    emit messageAvailable( "" );

}


//___________________________________________________________________________
void FigureFrame::_installActions( void )
{
    Debug::Throw( "FigureFrame::_installActions.\n" );

    // add buttons to menu
    addAction( viewFigureAction_ = new QAction( IconEngine::get( IconNames::Preview ), tr( "View Figure" ), this ) );
    connect( viewFigureAction_, SIGNAL(triggered()), SLOT(_viewFigure()) );

    addAction( reloadAction_ = new QAction( IconEngine::get( IconNames::Reload ), tr( "Reload" ), this ) );
    connect( reloadAction_, SIGNAL(triggered()), SLOT(requestUpdate()) );
    reloadAction_->setShortcut( QKeySequence::Refresh );
}
