/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "IncludeFrame.h"

#include "Command.h"
#include "ContextMenu.h"
#include "Debug.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "InformationDialog.h"
#include "TreeView.h"
#include "XmlOptions.h"

#include <QHeaderView>

//_____________________________________________________________
IncludeFrame::IncludeFrame( QTabWidget* parent ):
    TabWidget( parent )
{
    Debug::Throw("IncludeFrame::IncludeFrame.\n" );
    setTitle( tr( "Includes" ) );


    QBoxLayout* hLayout = new QHBoxLayout();
    hLayout->setSpacing(5);
    hLayout->setMargin(5);
    setLayout( hLayout );

    hLayout->addWidget( list_ = new TreeView( this ), 1 );
    model_.setSortValues( false );
    list_->setModel( &model_ );
    list_->setRootIsDecorated( true );
    list_->setSortingEnabled(false);
    list_->setSelectionMode( QAbstractItemView::ContiguousSelection );
    list_->header()->hide();

    _installActions();

    // add buttons to menu
    QMenu* menu( new ContextMenu( list_ ) );
    menu->addAction( editAction_ );
    menu->addAction( spellcheckAction_ );
    menu->addSeparator();
    menu->addAction( reloadAction_ );

    // buttons
    QVBoxLayout* vLayout( new QVBoxLayout() );
    vLayout->setMargin(0);
    vLayout->setSpacing(2);
    hLayout->addLayout( vLayout );

    vLayout->addWidget( buttonWidget_ = new QWidget( this ), 0 );
    {
        QVBoxLayout* vLayout( new QVBoxLayout() );
        vLayout->setMargin(0);
        vLayout->setSpacing(2);
        buttonWidget_->setLayout( vLayout );

        vLayout->addWidget( selectAllButton_ = new QPushButton( tr( "Select All" ), buttonWidget_ ) );
        vLayout->addWidget( selectNoneButton_ = new QPushButton( tr( "Select None" ), buttonWidget_ ) );
        vLayout->addWidget( invertButton_ = new QPushButton( tr( "Invert Selection" ), buttonWidget_ ) );

        // separator
        QFrame* frame( new QFrame( buttonWidget_ ) );
        frame->setFrameStyle( QFrame::HLine|QFrame::Sunken );
        vLayout->addWidget( frame );

    }

    vLayout->addWidget( editButton_ = new QPushButton( tr( "Edit" ), this ) );
    vLayout->addWidget( spellCheckButton_ = new QPushButton( tr( "Spellcheck" ), this ) );

    vLayout->addStretch();

    // connections
    connect( list_->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), SLOT(_updateActions()) );
    connect( list_, SIGNAL(activated(QModelIndex)), editAction_, SIGNAL(triggered()) );
    connect( list_, SIGNAL(clicked(QModelIndex)), SLOT(_toggle(QModelIndex)) );
    connect( list_, SIGNAL(clicked(QModelIndex)), SLOT(_updateButtons()) );
    connect( selectAllButton_, SIGNAL(pressed()), SLOT(_selectAll()) );
    connect( selectNoneButton_, SIGNAL(pressed()), SLOT(_selectNone()) );
    connect( invertButton_, SIGNAL(pressed()), SLOT(_invert()) );
    connect( editButton_, SIGNAL(pressed()), SLOT(_edit()) );
    connect( spellCheckButton_, SIGNAL(pressed()), SLOT(_spellCheck()) );
    connect( &fileSystemWatcher_, SIGNAL(fileChanged(QString)), SLOT(requestUpdate()) );
    buttonWidget_->setVisible( false );
    _updateActions();

}

//_____________________________________________________________
void IncludeFrame::setFile( const File& value )
{
    Debug::Throw() <<  "FigureFrame::setFile - " << value << endl;
    if( file_ != value )
    {
        file_ = value;
        requestUpdate();
    }
}

//_____________________________________________________________
bool IncludeFrame::filesModified( void )
{

    Debug::Throw( "IncludeFrame::filesModified.\n" );

    // check last updated
    if( !lastUpdate_.isValid() ) return true;

    // loop over include files
    for( const auto& include:model_.children() )
    {
        if( !_fileIsAfs( include.file() ) ) continue;
        if( !include.file().exists() ) continue;
        if( include.file().lastModified() > lastUpdate_ ) return true;
    }

    return false;

}

//_____________________________________________________________
IncludeFileModel::List IncludeFrame::selection( void ) const
{ return model_.get( list_->selectionModel()->selectedRows() ); }

//_____________________________________________________________
void IncludeFrame::requestUpdate( void )
{ emit commandRequested( LatexCommand( LatexCommand::FindIncludes ) << file_ ); }

//_____________________________________________________________
void IncludeFrame::update( const IncludeFile::List& includes )
{
    Debug::Throw( "IncludeFrame::Update.\n" );

    // copy old includes
    const IncludeFile::List oldIncludes( model_.children() );

    // update filesystem watcher
    // remove old files that are not in current any more
    for( const auto& includeFile:oldIncludes )
    {
        if( !includes.contains( includeFile ) )
        { fileSystemWatcher_.removePath( includeFile.file() ); }
    }

    // add new files
    for( const auto& includeFile:includes )
    {
        if( !oldIncludes.contains( includeFile ) && includeFile.file().exists() )
        { fileSystemWatcher_.addPath( includeFile.file() ); }
    }

    // model_.clear();
    model_.set( includes );

    // look for new items and expand parent automatically
    for( const auto& includeFile:includes )
    {
        if( !oldIncludes.contains( includeFile ) )
        { list_->expand( model_.parent( model_.index( includeFile ) ) ); }
    }

    // update buttons and resize list
    _updateButtons();
    list_->resizeColumns();

    // update timeStamp
    lastUpdate_ = TimeStamp::now();

    return;
}

//_____________________________________________________________
void IncludeFrame::_updateActions( void )
{
    Debug::Throw( "FigureFrame::_updateActions.\n" );
    bool hasSelection( !list_->selectionModel()->selectedRows().empty() );
    editAction_->setEnabled( hasSelection );
    spellcheckAction_->setEnabled( hasSelection );
    return;
}



//_____________________________________________________________
void IncludeFrame::_edit( void )
{

    // get selection
    Debug::Throw( "IncludeFrame::_edit.\n" );

    // retrieve selection
    auto selection( model_.get( list_->selectionModel()->selectedRows() ) );
    if( selection.empty() )
    {
        InformationDialog( this, tr( "No File selected. <Edit> canceled." ) ).exec();
        return;
    }

    // check editor
    if( !XmlOptions::get().contains( "TEXT_EDITOR" ) ) {
        InformationDialog( this, tr( "Text editor not set." ) ).exec();
        return;
    }

    Command command( XmlOptions::get().raw( "TEXT_EDITOR" ) );
    for( const auto& includeFile:selection )
    {

        const File& file( includeFile.file() );

        if( !file.exists() )
        {
            const QString buffer = QString( "File '%1' does not exist. <Edit> canceled." ).arg( file );
            InformationDialog( this, buffer ).exec();
            continue;
        }

        if( file.isDirectory() )
        {
            const QString buffer = QString( "File '%1' is a directory. <Edit> canceled." ).arg( file );
            InformationDialog( this, buffer ).exec();
            continue;
        }

        command << file;

    }

    emit messageAvailable( QString( tr( "Running %1..." ) ).arg( QString( XmlOptions::get().raw("TEXT_EDITOR" ) ) ) );

    command.run();

    emit messageAvailable( "" );
}

//_____________________________________________________________
void IncludeFrame::_spellCheck(void)
{
    Debug::Throw( "IncludeFrame::_SpellCheck.\n" );

    // retrieve selection
    auto selection( model_.get( list_->selectionModel()->selectedRows() ) );
    if( selection.empty() )
    {
        InformationDialog( this, tr( "No File selected. <Spellcheck> canceled." ) ).exec();
        return;
    }

    // check spell checker command
    if( !XmlOptions::get().contains( "SPELLCHECK" ) ) {
        InformationDialog( this, tr( "Spellcheck command not set." ) ).exec();
        return;
    }

    Command command( XmlOptions::get().raw( "SPELLCHECK" ) );
    for( const auto& includeFile:selection )
    {

        const File& file( includeFile.file() );
        if( !file.exists() )
        {
            const QString buffer = QString( "File '%1' does not exist. <Spellcheck> canceled." ).arg( file );
            InformationDialog( this, buffer ).exec();
            continue;
        }

        if( file.isDirectory() ) {
            const QString buffer = QString( "File '%1' is a directory. <Spellcheck> canceled." ).arg( file );
            InformationDialog( this, buffer ).exec();
            continue;
        }

        // update display and run.
        command << file;
    }

    emit messageAvailable( QString( tr( "Running %1..." ) ).arg( QString( XmlOptions::get().raw("SPELLCHECK" ) ) ) );
    command.run();

    emit messageAvailable( "" );

}

//_____________________________________________________________
void IncludeFrame::_selectAll(void)
{
    Debug::Throw( "IncludeFrame::_selectAll.\n" );

    // loop over files
    for( const auto& includeFile:model_.children() )
    {
        if( includeFile.parent().file().isEmpty() ) continue;
        if( !includeFile.isSelected() ) _toggle( model_.index( includeFile ), false );
    }

    selectAllButton_->setEnabled( false );
    selectNoneButton_->setEnabled( true );

}

//_____________________________________________________________
void IncludeFrame::_selectNone(void)
{
    Debug::Throw( "IncludeFrame::_selectNone.\n" );

    // get all files
    auto includeFiles( model_.children() );

    // loop over files
    for( const auto& includeFile:model_.children() )
    {
        if( includeFile.parent().file().isEmpty() ) continue;
        if( includeFile.isSelected() )
        { _toggle( model_.index( includeFile ), false ); }
    }

    selectAllButton_->setEnabled( true );
    selectNoneButton_->setEnabled( false );

}

//_____________________________________________________________
void IncludeFrame::_invert(void)
{
    Debug::Throw( "IncludeFrame::_invert.\n" );

    // loop over files
    for( const auto& includeFile:model_.children() )
    {
        if( !includeFile.parent().file().isEmpty() )
        { _toggle( model_.index( includeFile ), false ); }
    }

}

//_______________________________________________
void IncludeFrame::_toggle( const QModelIndex& index, bool checkColumn )
{
    Debug::Throw( "IncludeFrame::_toggle.\n" );

    // check index validity, parent, and column
    if( !index.isValid() ) return;
    if( !model_.parent( index ).isValid() ) return;
    if( checkColumn && index.column() != IncludeFileModel::Selection ) return;

    // convert to File
    IncludeFile includeFile( model_.get( index ) );

    // get file and parent
    File thisFile( includeFile.file() );
    File parentFile( model_.get( model_.parent( index ) ).file() );

    // emit command
    emit commandRequested( LatexCommand( LatexCommand::ToggleIncludeFile ) << parentFile << thisFile );

    return;
}

//______________________________________________________________________
void IncludeFrame::_updateButtons( void )
{
    Debug::Throw( "IncludeFrame::_updateButtons.\n" );

    bool hasCheckable( false );
    bool hasChecked( false );
    bool hasUnchecked( false );

    // loop over files
    for( const auto& includeFile:model_.children() )
    {
        if( includeFile.parent().file().isEmpty() ) continue;

        hasCheckable = true;
        if( includeFile.isSelected() ) hasChecked = true;
        else hasUnchecked = true;

    }

    selectAllButton_->setEnabled( hasUnchecked );
    selectNoneButton_->setEnabled( hasChecked );
    invertButton_->setEnabled( hasCheckable );

    buttonWidget_->setVisible( hasCheckable );

    // selection
    bool hasSelection( !list_->selectionModel()->selectedRows().isEmpty() );
    editButton_->setEnabled( hasSelection );
    spellCheckButton_->setEnabled( hasSelection );

}

//___________________________________________________________________________
bool IncludeFrame::_fileIsAfs( const QString& file ) const
{ return file.indexOf( "/afs" ) == 0; }

//___________________________________________________________________________
void IncludeFrame::_installActions( void )
{
    Debug::Throw( "IncludeFrame::_installActions.\n" );
    addAction( editAction_ = new QAction( IconEngine::get( IconNames::Edit ), tr( "Edit" ), this ) );
    connect( editAction_, SIGNAL(triggered()), SLOT(_edit()) );

    addAction( spellcheckAction_ = new QAction( IconEngine::get( IconNames::SpellCheck ), tr( "Check Spelling" ), this ) );
    connect( spellcheckAction_, SIGNAL(triggered()), SLOT(_spellCheck()) );

    addAction( reloadAction_ = new QAction( IconEngine::get( IconNames::Reload ), tr( "Reload" ), this ) );
    connect( reloadAction_, SIGNAL(triggered()), SLOT(requestUpdate()) );
    reloadAction_->setShortcut( QKeySequence::Refresh );

}
