#ifndef LatexCommand_h
#define LatexCommand_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Counter.h"
#include "Debug.h"

#include <QString>
#include <QStringList>
#include <QHash>
#include <QList>

//* handles top commands
class LatexCommand: public Counter
{

    public:

    //* enumeration for command type
    enum Type
    {
        None,
        Latex,
        PDFLatex,
        Bibtex,
        DviPS,
        DviPDF,
        PS2PDF,
        Check,
        CheckCitations,
        CheckCitationOrder,
        CheckLabels,
        CheckReferences,
        FindIncludes,
        FindFigures,
        ToggleIncludeFile,
        Clean

    };

    //* list of commands
    using List = QList<LatexCommand>;

    //* constructor
    LatexCommand( const Type = None );

    //* less than operator
    virtual bool operator < ( const LatexCommand& other ) const
    { return id() < other.id(); }

    //* equal to operator
    bool operator == (const LatexCommand& other ) const
    { return id() == other.id(); }

    //*@name accessors
    //@{

    //* command name
    QString name( void ) const
    { return name( type() ); }

    //* map Type to command name
    static QString name( const Type& type );

    //* id
    int id( void ) const
    { return id_; }

    //* type
    Type type( void ) const
    { return type_; }

    // true if arguments are passed
    bool hasArguments( void ) const
    { return !arguments_.isEmpty(); }

    //* arguments
    const QStringList& arguments( void ) const
    { return arguments_; }

    //* path
    QString path( void ) const;

    //@}

    //*@name modifiers
    //@{

    //* add argument
    LatexCommand& addArgument( const QString& argument )
    {
        arguments_ << argument;
        return *this;
    }

    //* push arguments
    LatexCommand& operator << (const QString& argument )
    { return addArgument( argument ); }

    //* clear arguments
    void clearArguments( void )
    { arguments_.clear(); }

    //@}

    private:

    //* static id counter
    static unsigned int& _counter( void );

    //* map types to command name
    using NameMap = QHash<Type,QString>;

    //* name map
    static NameMap& _names( void );

    //* unique id
    int id_;

    //* type
    Type type_;

    //* arguments
    QStringList arguments_;

    //* initialize command names
    static void _initializeNames( void );

};

#endif
