#ifndef LatexInterface_h
#define LatexInterface_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Debug.h"
#include "File.h"
#include "IncludeFile.h"
#include "InformationDialog.h"
#include "LatexCommand.h"
#include "MessageBuffer.h"
#include "TextFormat.h"

#include <QObject>
#include <QProcess>
#include <QString>

class CustomProcess;

//* Job managment
class LatexInterface:public QObject, public Counter
{

    //* Qt meta object declaration
    Q_OBJECT

    public:

    //* constructor
    LatexInterface( QObject* parent = nullptr );

    //* check file existence
    bool checkFile( File file, bool quiet = false )
    { return _checkFile( file, QString(), quiet ); }

    Q_SIGNALS:

    //* command added
    void commandAdded( const LatexCommand& );

    //* commands added
    void commandsAdded( const LatexCommand::List& );

    //* command completed
    void commandCompleted( const LatexCommand& );

    //* command started
    void commandStarted( const LatexCommand& );

    //* emitted when executing a command
    void busy( void );

    //* emitted when there is no more commands in pile
    void ready( void );

    //* emitted when busy state changed
    void busyStateChanged( bool );

    //* logAvailable
    void logAvailable( const QString&, Format::TextFormatFlags = Format::Default );

    //* logAvailable
    void warningLogAvailable( const QString&, Format::TextFormatFlags = Format::Default );

    //* warnings
    void warningAvailable( const QString& );

    //* message
    void messageAvailable( const QString& );

    //* include files available
    void includeFilesAvailable( const IncludeFile::List& );

    //* bibliography file
    void bibtexFileAvailable( const File& );

    //* graphics path available
    void graphicsPathAvailable( const File& );

    //* figure list available
    void figuresAvailable( const FileRecord::List& );

    //* current page changed
    void pageChanged( int );

    //* max page change
    void maxPageChanged( int );

    public Q_SLOTS:

    //* add command
    void addCommand( const LatexCommand& );

    //* add commands
    void addCommands( const LatexCommand::List& );

    //* abort command
    void abortCommands( bool quiet = false );

    protected Q_SLOTS:

    //* error from process
    void _error( QProcess::ProcessError );

    //* take action when current command completed
    void _commandCompleted( int, QProcess::ExitStatus );

    //* read on-fly from process
    void _readStdout( void );

    //* read on-fly from process
    void _readStderr( void );

    protected:

    //* check file existence
    bool _checkFile( File, const QString&, bool quiet = false );

    //* reset buffers
    void _resetBuffers( void );

    //* run next command in pile
    bool _runCommand( void );

    //* check references
    void _checkReferences( const File& );

    //* check citations
    void _checkCitations( const File& );

    //* check labels
    void _checkLabels( const File& );

    //* check citations order
    void _checkCitationOrder( const File& );

    //* find includes (recursive)
    void _findIncludes( const IncludeFile&, IncludeFile::List& );

    //* toggle include file
    void _toggleIncludeFile( const File&, const File& ) const;

    //* find graphics path
    void _findFigures( const File& );

    //* clean
    void _clean( const File&, bool& first );

    //* clean
    void _clean( const File&, const QString&, bool& first );

    //* clear process
    void _resetProcess( void );

    //* create new process to run command
    CustomProcess& _newProcess( void );

    //* process
    CustomProcess& _process( void ) const
    { return *process_; }

    private:

    //* custom process
    CustomProcess* process_ = nullptr;

    //* stdout buffer
    MessageBuffer stdoutBuffer_;

    //* stderr buffer
    MessageBuffer stderrBuffer_;

    //* map command type to name
    LatexCommand::List commands_;

    //* cached list of include files
    IncludeFile::List includeFiles_;

    //* page number
    int currentPage_ = 0;

};

#endif
