#ifndef MainWindow_h
#define MainWindow_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "LineEditor.h"
#include "BaseMainWindow.h"
#include "Counter.h"
#include "Debug.h"
#include "FileRecord.h"
#include "LatexCommand.h"
#include "WindowTitle.h"

#include <QAction>
#include <QEvent>
#include <QTabWidget>

class ActionFrame;
class ButtonFrame;
class FigureFrame;
class IncludeFrame;
class LatexInterface;
class LogFrame;
class Menu;
class RecentFilesFrame;
class StatusBar;

/*!
\class MainWindow
\brief customized QVBox to recieve close event
*/
class MainWindow: public BaseMainWindow
{

    //* Qt meta object declaration
    Q_OBJECT

    public:

    //* constructor;
    MainWindow( QWidget* );

    //*@name accessors
    //@{

    //* retrieve Menu singleton
    Menu &menu( void ) const
    { return *menu_; }

    //* retrieve ActionFrame
    ActionFrame &actionFrame( void ) const
    { return *actionFrame_; }

    //* retrieve ButtonFrame
    ButtonFrame &buttonFrame( void ) const
    { return *buttonFrame_; }

    //* retrieve LogFrame
    LogFrame &logFrame( void ) const
    { return *logFrame_; }

    //* retrieve IncludeFrame
    IncludeFrame &includeFrame( void ) const
    { return *includeFrame_; }

    //* retrieve FigureFrame
    FigureFrame &figureFrame( void ) const
    { return *figureFrame_; }

    //* retrieve recent file frame
    RecentFilesFrame &recentFilesFrame( void ) const
    { return *recentFilesFrame_; }

    //* retrieve StatusBar
    StatusBar& statusBar( void ) const
    { return *statusbar_; }

    //* latex interface
    LatexInterface& interface( void ) const
    { return *interface_; }

    //* retrieve file from editor
    File file( void ) const
    { return editor_->text(); }

    //* retrieve file from editor
    bool hasBibtexFile( void ) const
    { return !bibtexFile_.isEmpty() && bibtexFile_.exists(); }

    //* bibtex file
    const File& bibtexFile( void ) const
    { return bibtexFile_; }

    //@}

    //*@name modifiers
    //@{

    //* set file
    void setFile( const File& file );

    //* check file
    bool checkFile( const File&, bool quiet = false );

    //*@name actions
    //@{

    //* open action
    QAction& openAction( void ) const
    { return *openAction_; }

    //* to latex accent
    QAction& toLatexAccentsAction( void ) const
    { return *toLatexAccentsAction_; }

    //* to text accents
    QAction& toTextAccentsAction( void ) const
    { return *toTextAccentsAction_; }

    //@}

    public Q_SLOTS:

    //* edit file (with extension if any, row otherwise)
    void edit( const QString& = QString() );

    //* edit tex file
    void open( FileRecord = FileRecord() );

    //* edit tex file
    void editTex( void )
    { edit( ".tex" ); }

    //* edit log file
    void editLog( void )
    { edit( ".log" ); }

    //* edit aux file
    void editAux( void )
    { edit( ".aux" ); }

    //* edit Toc file
    void editToc( void )
    { edit( ".toc" ); }

    //* edit Bib file
    void editBib( void )
    { if( hasBibtexFile() ) _edit( bibtexFile_ ); }

    //* edit Tof file
    void editTof( void )
    { edit( ".tof" ); }

    //* edit ps file
    void editPS( void )
    { edit( ".ps" ); }

    //* add latex header/trailer to text file
    void addLatexHeader( void );

    Q_SIGNALS:

    void resizeRequest( void );

    protected:

    //* close window event handler
    virtual void closeEvent( QCloseEvent* );

    //* enter
    virtual void enterEvent( QEvent* );

    //* edit
    virtual bool _edit( const File& );

    protected Q_SLOTS:

    //* command has been completed
    void _processCompletedCommand( const LatexCommand& );

    //* update file from text editor
    void _updateFile( void );

    //* update window title
    void _updateWindowTitle( const QString& text = QString() )
    { setWindowTitle( WindowTitle( text ) ); }

    //* replace accents (text to latex)
    void _toLatexAccents( void );

    //* replace accents (latex to text)
    void _toTextAccents( void );

    //* bibtex file available
    void _updateBibtexFile( const File& );

    private Q_SLOTS:

    //* update configuration
    void _updateConfiguration();

    private:

    //* install actions
    void _installActions( void );

    //* needed to keep track of actual file changes
    File previousFile_;

    //* bibliography file
    File bibtexFile_;

    //* latex interface
    LatexInterface* interface_ = nullptr;

    //* top widget
    QWidget* top_ = nullptr;

    //* file line editor singleton
    LineEditor* editor_ = nullptr;

    //* action frame singleton
    ActionFrame* actionFrame_ = nullptr;

    //* button frame singleton
    ButtonFrame* buttonFrame_ = nullptr;

    //*@name actions
    //@{

    //* open
    QAction* openAction_ = nullptr;

    //* to latex accents
    QAction* toLatexAccentsAction_ = nullptr;

    //* to text accents
    QAction* toTextAccentsAction_ = nullptr;

    //@}

    //* tab widget
    QTabWidget *tabWidget_ = nullptr;

    //* tab widget size
    int tabWidgetHeight_ = 0;

    //* log frame singleton
    LogFrame* logFrame_ = nullptr;

    //* includes frame singleton
    IncludeFrame* includeFrame_ = nullptr;

    //* figures frame singleton
    FigureFrame* figureFrame_ = nullptr;

    //* recent files frame
    RecentFilesFrame* recentFilesFrame_ = nullptr;

    //* state frame singleton
    StatusBar* statusbar_ = nullptr;

    //* Menu singleton
    Menu* menu_ = nullptr;

};

#endif
